/**
 * 
 */
package org.gcube.informationsystem.resourceregistry.instances.type;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.activation.UnsupportedDataTypeException;

import org.gcube.informationsystem.base.reference.AccessType;
import org.gcube.informationsystem.base.reference.ISManageable;
import org.gcube.informationsystem.base.reference.entities.BaseEntity;
import org.gcube.informationsystem.base.reference.properties.BaseProperty;
import org.gcube.informationsystem.base.reference.properties.Header;
import org.gcube.informationsystem.base.reference.relations.BaseRelation;
import org.gcube.informationsystem.model.reference.entities.Entity;
import org.gcube.informationsystem.model.reference.entities.Resource;
import org.gcube.informationsystem.model.reference.properties.Property;
import org.gcube.informationsystem.model.reference.relations.Relation;
import org.gcube.informationsystem.resourceregistry.api.exceptions.schema.SchemaAlreadyPresentException;
import org.gcube.informationsystem.resourceregistry.api.exceptions.schema.SchemaCreationException;
import org.gcube.informationsystem.resourceregistry.api.exceptions.schema.SchemaException;
import org.gcube.informationsystem.resourceregistry.api.exceptions.schema.SchemaNotFoundException;
import org.gcube.informationsystem.resourceregistry.instances.base.ERManagement;
import org.gcube.informationsystem.resourceregistry.instances.context.ContextUtility;
import org.gcube.informationsystem.resourceregistry.instances.type.entities.EntityTypeDefinitionManagement;
import org.gcube.informationsystem.resourceregistry.instances.type.properties.PropertyTypeDefinitionManagement;
import org.gcube.informationsystem.resourceregistry.instances.type.relations.RelationTypeDefinitionManagement;
import org.gcube.informationsystem.resourceregistry.security.AdminSecurityContext;
import org.gcube.informationsystem.resourceregistry.security.SecurityContext.PermissionMode;
import org.gcube.informationsystem.types.Type;
import org.gcube.informationsystem.types.TypeBinder;
import org.gcube.informationsystem.types.reference.TypeDefinition;
import org.gcube.informationsystem.types.reference.entities.EntityTypeDefinition;
import org.gcube.informationsystem.types.reference.properties.PropertyDefinition;
import org.gcube.informationsystem.types.reference.properties.PropertyTypeDefinition;
import org.gcube.informationsystem.types.reference.relations.RelationTypeDefinition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.orientechnologies.orient.core.db.document.ODatabaseDocument;
import com.orientechnologies.orient.core.exception.OSchemaException;
import com.orientechnologies.orient.core.metadata.OMetadata;
import com.orientechnologies.orient.core.metadata.schema.OClass;
import com.orientechnologies.orient.core.metadata.schema.OClassImpl;
import com.orientechnologies.orient.core.metadata.schema.OProperty;
import com.orientechnologies.orient.core.metadata.schema.OSchema;
import com.orientechnologies.orient.core.metadata.schema.OType;
import com.orientechnologies.orient.core.record.OElement;
import com.orientechnologies.orient.core.record.impl.ODocument;

/**
 * @author Luca Frosini (ISTI - CNR)
 */
public class SchemaManagementImpl implements SchemaManagement {
	
	private static Logger logger = LoggerFactory.getLogger(SchemaManagementImpl.class);
	
	protected String typeName;
	
	protected OClass getOClass(OSchema oSchema, String type) throws SchemaException {
		return oSchema.getClass(type);
	}
	
	public void setTypeName(String typeName) {
		this.typeName = typeName;
	}
	
	@SuppressWarnings("unused")
	private static TypeDefinition getOClassTypeDefinition(OClass oClass) throws SchemaException {
		try {
			ODocument oDocument = ((OClassImpl) oClass).toStream();
			String json = oDocument.toJSON();
			ObjectMapper mapper = new ObjectMapper();
			ObjectNode node = (ObjectNode) mapper.readTree(json);
			
			if(oClass.isSubClassOf(BaseProperty.NAME)) {
				node.put(ISManageable.CLASS_PROPERTY, PropertyTypeDefinition.NAME);
			} else if(oClass.isSubClassOf(BaseEntity.NAME)) {
				node.put(ISManageable.CLASS_PROPERTY, EntityTypeDefinition.NAME);
			} else if(oClass.isSubClassOf(BaseRelation.NAME)) {
				node.put(ISManageable.CLASS_PROPERTY, RelationTypeDefinition.NAME);
			}
			
			ArrayNode arrayNode = (ArrayNode) node.get(TypeDefinition.PROPERTIES_PROPERTY);
			Iterator<JsonNode> iterator = arrayNode.iterator();
			while(iterator.hasNext()) {
				ObjectNode propertyNode = (ObjectNode) iterator.next();
				propertyNode.put(ISManageable.CLASS_PROPERTY, PropertyDefinition.NAME);
			}
			
			String managedJson = mapper.writeValueAsString(node);
			logger.trace("{} -> {}", json, managedJson);
			
			return TypeBinder.deserializeTypeDefinition(managedJson);
		} catch(Exception e) {
			throw new SchemaException(e);
		}
	}
	
	private TypeDefinition getTypeDefinition(OClass oClass) throws SchemaException {
		try {
			ERManagement<? extends OElement> erManagement = null;
			
			if(oClass.isSubClassOf(BaseProperty.NAME)) {
				erManagement = new PropertyTypeDefinitionManagement();
				((PropertyTypeDefinitionManagement) erManagement).setName(oClass.getName());
			} else if(oClass.isSubClassOf(BaseEntity.NAME)) {
				erManagement = new EntityTypeDefinitionManagement();
				((EntityTypeDefinitionManagement) erManagement).setName(oClass.getName());
			} else if(oClass.isSubClassOf(BaseRelation.NAME)) {
				erManagement = new RelationTypeDefinitionManagement();
				((RelationTypeDefinitionManagement) erManagement).setName(oClass.getName());
			}
			
			String ret = erManagement.read();
			return TypeBinder.deserializeTypeDefinition(ret);
		} catch(Exception e) {
			throw new SchemaException(e);
		}
	}
	
	protected String getTypeDefinitionAsString(OClass oClass) throws SchemaException {
		try {
			TypeDefinition typeDefinition = getTypeDefinition(oClass);
			return TypeBinder.serializeTypeDefinition(typeDefinition);
		} catch(Exception e) {
			throw new SchemaException(e);
		}
	}
	
	protected List<OClass> getSuperclassesAndCheckCompliancy(ODatabaseDocument oDatabaseDocument,
			TypeDefinition typeDefinition, String baseType) throws SchemaException, SchemaNotFoundException {
		
		Set<String> superClasses = typeDefinition.getSuperClasses();
		if(baseType != null) {
			if(superClasses == null || superClasses.size() == 0) {
				throw new RuntimeException(
						String.format("No Superclass found in schema %s. The Type Definition must extend %s",
								typeDefinition, baseType));
			}
		}
		
		OMetadata oMetadata = oDatabaseDocument.getMetadata();
		OSchema oSchema = oMetadata.getSchema();
		
		List<OClass> oSuperclasses = new ArrayList<>();
		for(String superClass : superClasses) {
			OClass oSuperClass = getOClass(oSchema, superClass);
			if(oSuperClass == null) {
				throw new SchemaNotFoundException("Superclass " + superClass + " does not exists");
			}
			if(baseType != null) {
				if(typeDefinition.getName().compareTo(baseType) != 0) {
					if(!oSuperClass.isSubClassOf(baseType)) {
						throw new RuntimeException(superClass + " is not a subsclass of " + baseType
								+ ". Each Superclass MUST be a subclass of " + baseType);
					}
				}
			}
			oSuperclasses.add(oSuperClass);
		}
		
		return oSuperclasses;
	}
	
	private static Set<String> baseTypes;
	public static Set<String> typeDefinitionTypes;
	public static Set<String> skipTypeDefinitionCreation;
	
	static {
		baseTypes = new HashSet<String>();
		baseTypes.add(BaseProperty.NAME);
		baseTypes.add(BaseEntity.NAME);
		baseTypes.add(BaseRelation.NAME);
		
		typeDefinitionTypes = new HashSet<String>();
		typeDefinitionTypes.add(PropertyTypeDefinition.NAME);
		typeDefinitionTypes.add(EntityTypeDefinition.NAME);
		typeDefinitionTypes.add(RelationTypeDefinition.NAME);
		
		skipTypeDefinitionCreation = new HashSet<String>();
		skipTypeDefinitionCreation.addAll(baseTypes);
		skipTypeDefinitionCreation.add(Header.NAME);
		skipTypeDefinitionCreation.add(PropertyDefinition.NAME);
	}
	
	protected void registerTypeSchema(TypeDefinition typeDefinition, AccessType baseType)
			throws SchemaAlreadyPresentException, SchemaException {
		
		ODatabaseDocument oDatabaseDocument = null;
		try {
			
			if(typeName.compareTo(typeDefinition.getName()) != 0) {
				String error = String.format(
						"Provided type name path argument %s does not match with the type name in the definition %S. Please be coherent.",
						typeName, typeDefinition.getName());
				throw new SchemaCreationException(error);
			}
			
			AdminSecurityContext adminSecurityContext = ContextUtility.getAdminSecurityContext();
			oDatabaseDocument = adminSecurityContext.getDatabaseDocument(PermissionMode.WRITER);
			
			OMetadata oMetadata = oDatabaseDocument.getMetadata();
			OSchema oSchema = oMetadata.getSchema();
			
			OClass oClass = null;
			
			if(BaseEntity.class.isAssignableFrom(baseType.getTypeClass())) {
				oClass = oDatabaseDocument.createVertexClass(typeDefinition.getName());
			} else if(BaseRelation.class.isAssignableFrom(baseType.getTypeClass())) {
				oClass = oDatabaseDocument.createEdgeClass(typeDefinition.getName());
			} else if(BaseProperty.class.isAssignableFrom(baseType.getTypeClass())) {
				oClass = oSchema.createClass(typeDefinition.getName());
			} else {
				String error = String.format("Allowed superclass are %s, %s, %s, or any subclasses of them.",
						Entity.NAME, Relation.NAME, Property.NAME);
				throw new SchemaCreationException(error);
			}
			
			try {
				
				String description = typeDefinition.getDescription();
				if(description != null && description.compareTo("") != 0) {
					try {
						oClass.setDescription(description);
					} catch(Exception e) {
						logger.warn(
								"Unable to set description. This is an orient bug. See https://github.com/orientechnologies/orientdb/issues/7065");
					}
				}
				
				try {
					// oClass.setAbstract(false); // Used to allow to persist Schema in Context
					// Management
					oClass.setAbstract(typeDefinition.isAbstract());
				} catch(Exception e) {
					logger.error(
							"Unable to set the Vertex Type {} as abstract. This is an OrientDB <= 2.2.12 bug. The Type will be created as it is not abstract.",
							typeDefinition.getName());
				}
				
				if(!baseTypes.contains(typeDefinition.getName())) {
					List<OClass> oSuperclasses = getSuperclassesAndCheckCompliancy(oDatabaseDocument, typeDefinition,
							baseType.getName());
					oClass.setSuperClasses(oSuperclasses);
				}
				
				Set<PropertyDefinition> propertyDefinitions = typeDefinition.getProperties();
				if(Resource.class.isAssignableFrom(baseType.getTypeClass())) {
					if(propertyDefinitions != null && propertyDefinitions.size() > 0) {
						throw new SchemaCreationException("A Resource cannot contains any properties.");
					}
				} else {
					for(PropertyDefinition propertyDefinition : propertyDefinitions) {
						
						OType oType = OType.getById(propertyDefinition.getType().byteValue());
						
						/*
						 * Types update is not allowed, 
						 * hence bug https://github.com/orientechnologies/orientdb/issues/7354 cannot occur
						 * Excluding the check from types used for type definition
						 * 
						 */
						if(!typeDefinitionTypes.contains(typeDefinition.getName())) {
							switch(oType) {
								case EMBEDDEDLIST:
									throw new UnsupportedDataTypeException(Type.OType.PROPERTYLIST
											+ " support is currently disabled due to OrientDB bug see https://github.com/orientechnologies/orientdb/issues/7354");
								case EMBEDDEDSET:
									throw new UnsupportedDataTypeException(Type.OType.PROPERTYSET
											+ " support is currently disabled due to OrientDB bug see https://github.com/orientechnologies/orientdb/issues/7354");
								default:
									break;
							}
						}
						
						OProperty op = oClass.createProperty(propertyDefinition.getName(), oType);
						op.setDescription(propertyDefinition.getDescription());
						
						/*
						 * Mandatory and notNull does not work in distributed mode: so that on Type
						 * declaration they are forced to false
						 * ovp.setMandatory(property.isMandatory());
						 * ovp.setNotNull(property.isNotnull()); This information are persisted in
						 * Management Context
						 */
						op.setMandatory(false);
						op.setNotNull(false);
						
						op.setReadonly(propertyDefinition.isReadonly());
						op.setRegexp(propertyDefinition.getRegexp());
						
						if(propertyDefinition.getLinkedClass() != null) {
							OClass linkedClass = getOClass(oSchema, propertyDefinition.getLinkedClass());
							if(linkedClass == null) {
								logger.trace("class {} not found in schema", propertyDefinition.getLinkedClass());
								throw new Exception(
										"class " + propertyDefinition.getLinkedClass() + " not found in schema");
							}
							
							if(linkedClass.isEdgeType() || linkedClass.isVertexType()) {
								throw new Exception("A Property Field cannot be an Entity or a Relation");
							}
							
							op.setLinkedClass(linkedClass);
						} else if(propertyDefinition.getLinkedType() != null) {
							op.setLinkedType(OType.getById(propertyDefinition.getLinkedType().byteValue()));
						}
					}
				}
				
				oDatabaseDocument.commit();
				
				logger.info("{} {} registered successfully", baseType.getName(), typeDefinition.getName());
			} catch(Exception e) {
				oSchema.dropClass(typeDefinition.getName());
				throw e;
			}
		} catch(OSchemaException ex) {
			if(ex.getMessage().contains("already exists")) {
				throw new SchemaAlreadyPresentException(ex);
			}
			throw new SchemaException(ex);
		} catch(SchemaException e) {
			throw e;
		} catch(Exception ex) {
			throw new SchemaCreationException(ex);
		} finally {
			oDatabaseDocument.close();
		}
	}
	
	protected String getSchema(String type, boolean includeSubtypes) throws SchemaNotFoundException, SchemaException {
		ODatabaseDocument oDatabaseDocument = null;
		try {
			AdminSecurityContext adminSecurityContext = ContextUtility.getAdminSecurityContext();
			oDatabaseDocument = adminSecurityContext.getDatabaseDocument(PermissionMode.READER);
			
			OMetadata oMetadata = oDatabaseDocument.getMetadata();
			OSchema oSchema = oMetadata.getSchema();
			OClass baseOClass = oSchema.getClass(type);
			
			List<TypeDefinition> typeDefinitions = new ArrayList<>();
			typeDefinitions.add(getTypeDefinition(baseOClass));
			
			if(includeSubtypes) {
				Collection<OClass> subClasses = baseOClass.getAllSubclasses();
				for(OClass oClass : subClasses) {
					typeDefinitions.add(getTypeDefinition(oClass));
				}
			}
			
			return TypeBinder.serializeTypeDefinitions(typeDefinitions);
		} catch(SchemaException e) {
			throw e;
		} catch(Exception e) {
			throw new SchemaException(e);
		} finally {
			if(oDatabaseDocument != null) {
				oDatabaseDocument.close();
			}
		}
		
	}
	
	@Override
	public String create(String jsonSchema, AccessType accessType) throws SchemaAlreadyPresentException, SchemaException {
		TypeDefinition typeDefinition = null;
		try {
			try {
				typeDefinition = TypeBinder.deserializeTypeDefinition(jsonSchema);
				logger.info("Trying to register {} {} : {}", accessType.getName(), typeDefinition.getName(),
						jsonSchema);
			} catch(Exception e) {
				logger.error("Error while trying to register {} {}", accessType.getName(), jsonSchema);
				throw new SchemaCreationException(e);
			}
			
			registerTypeSchema(typeDefinition, accessType);
			
			ERManagement<? extends OElement> erManagement = null;
			
			if(BaseEntity.class.isAssignableFrom(accessType.getTypeClass())) {
				erManagement = new EntityTypeDefinitionManagement();
			} else if(BaseRelation.class.isAssignableFrom(accessType.getTypeClass())) {
				erManagement = new RelationTypeDefinitionManagement();
			} else if(BaseProperty.class.isAssignableFrom(accessType.getTypeClass())) {
				erManagement = new PropertyTypeDefinitionManagement();
			}
			
			String ret = null;
			if(!typeDefinitionTypes.contains(typeDefinition.getName())
					&& !skipTypeDefinitionCreation.contains(typeDefinition.getName())) {
				erManagement.setJson(jsonSchema);
				ret = erManagement.create();
			}else {
				ret = TypeBinder.serializeTypeDefinition(typeDefinition);
			}
			return ret;
		} catch(SchemaAlreadyPresentException e) {
			throw e;
		} catch(SchemaException e) {
			throw e;
		} catch(Exception ex) {
			throw new SchemaCreationException(ex);
		}
	}
	
	@Override
	public String read(String entityType, boolean includeSubtypes) throws SchemaNotFoundException, SchemaException {
		return getSchema(entityType, includeSubtypes);
	}
	
	@Override
	public String update(String entityType, AccessType accessType, String jsonSchema)
			throws SchemaNotFoundException, SchemaException {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public String delete(String entityType, AccessType accessType) throws SchemaNotFoundException {
		throw new UnsupportedOperationException();
	}
	
}
